﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Media.Media3D;
using System.Windows.Shapes;
using SciChart.Charting3D;
using SciChart.Charting3D.Model;
using SciChart.Charting3D.Model.ChartData;
using SciChart.Charting3D.Modifiers;
using SciChart.Charting3D.Primitives;
using SciChart.Charting3D.RenderableSeries;
using SciChart.Core.Extensions;
using SciChart.Core.Utility.Mouse;
using SciChart.Drawing.Utility;
using SelectionMode = SciChart.Charting.ChartModifiers.SelectionMode;

namespace DataPointSelectionModifier
{
  internal class DataPointSelectionModifier3D : VertexSelectionModifier3D
  {
    private readonly Dictionary<uint, HashSet<ulong>> mSelectedVertices = new Dictionary<uint, HashSet<ulong>>();
    private readonly Rectangle mPointRectangle = new Rectangle() {Stroke = Brushes.LimeGreen, StrokeThickness = 1, Fill = Brushes.LimeGreen, Opacity = 0.5};
    private readonly double mRectSizeFromCenter = 30;

    protected override SelectionMode GetSelectionMode(MouseModifier modifierKey, bool isAreaSelection)
    {
      SelectionMode selectionMode = base.GetSelectionMode(modifierKey, isAreaSelection);

      return selectionMode;
    }

    public override void OnModifierMouseDown(ModifierMouseArgs e)
    {
      if (e.Modifier != MouseModifier.Ctrl)
      {
        var rect = e.MousePoint.ToRectCenter(mRectSizeFromCenter);
        mPointRectangle.Height = rect.Height;
        mPointRectangle.Width = rect.Width;
        Canvas.SetTop(mPointRectangle, rect.Y);
        Canvas.SetLeft(mPointRectangle, rect.X);
        this.ModifierSurface.Children.AddIfNotContains(mPointRectangle);
      }
      
      base.OnModifierMouseDown(e);

    }

    public override void OnModifierMouseUp(ModifierMouseArgs mouseArgs)
    {
      ModifierSurface.Children.RemoveIfContains(mPointRectangle);
      
      //reset all inScene
      ResetMetadataInScene();

      if (mouseArgs.Modifier == MouseModifier.Ctrl)
      {
        base.OnModifierMouseUp(mouseArgs);
        return;
      }
      
      Point mousePoint = mouseArgs.MousePoint;
      var translatedPoint = this.RootGrid.TranslatePoint(mousePoint, this.ModifierSurface);
      
      List<EntityVertexId>? entityVertexIds = Viewport3D.PickScene(mousePoint.ToRectCenter(mRectSizeFromCenter));
      
      double distance = double.NaN;
      SeriesInfo3D? nearestSeriesInfo3D = null;
      if (entityVertexIds.Count > 0)
      {
        foreach (var entityVertexId in entityVertexIds)
        {
          Viewport3D.RootEntity.VisitEntities(entity =>
          {
            if (entity is Points3DSceneEntity pointsEntity)
            {
              if (entity.EntityId == entityVertexId.EntityId)
              {
                HitTestInfo3D hitTestInfo = new HitTestInfo3D()
                  {IsHit = true, EntityId = entityVertexId.EntityId, VertexId = entityVertexId.VertexId };
                SeriesInfo3D? seriesInfo = pointsEntity.ToSeriesInfo(hitTestInfo);

                //Get WorldCoordinates
                Vector3 vertex = seriesInfo.HitVertexCoords;

                //Convert vertex to 2D screen space
                Point screenPoint = ParentSurface.Camera.WorldToScreenSpace(vertex);
                //screenPoint.Y = ParentSurface.RootGrid.ActualHeight - screenPoint.Y; // not precisely

                //Fill additionalMetadata
                if (seriesInfo.PointMetadata is PointMetadata3DEx locatedMetadata)
                {
                  locatedMetadata.IsInScene = true;
                  locatedMetadata.ScreenPoint = screenPoint;
                  locatedMetadata.LastMousePoint = mousePoint;
                  locatedMetadata.MousePointTranslated = translatedPoint;
                }

                //determine point closest to mouse point
                double newDistance = PointUtil.Distance(screenPoint, translatedPoint);
                if (distance.IsNaN() || newDistance < distance)
                {
                  distance = newDistance;
                  if (nearestSeriesInfo3D != null)
                  {
                    nearestSeriesInfo3D.IsHit = false;
                  }
                  nearestSeriesInfo3D = seriesInfo;
                }
              }
            }
          });
        }

        IXyzDataSeries3D? selectedDataSeries = nearestSeriesInfo3D?.RenderableSeries?.DataSeries as IXyzDataSeries3D;
        if (selectedDataSeries == null)
        {
          return;
        }

        //if selection or focusing not possible - return
        if (!(nearestSeriesInfo3D?.PointMetadata is PointMetadata3DEx currentHitMetadata))
        {
          return;
        }

        nearestSeriesInfo3D.IsHit = true;
        currentHitMetadata.IsSelected = !currentHitMetadata.IsSelected;
      }
    }

    private void ResetMetadataInScene()
    {
      Viewport3D.RootEntity.VisitEntities(entity =>
      {
        if (entity is Points3DSceneEntity pointsEntity)
        {
          var seriesInfo = pointsEntity.ToSeriesInfo(new HitTestInfo3D());
          if (seriesInfo?.RenderableSeries.DataSeries is IXyzDataSeries3D dataSeries)
          {
            foreach (var pointMetadata3DEx in dataSeries.GetInSceneMetadata(true))
            {
              pointMetadata3DEx.IsInScene = false;
              pointMetadata3DEx.IsSelected = false;
            }
          }
        }
      });
    }
  }
}
